/*
   This file serves as an interface between TINKER Fortran code
   and FFE Java objects used for socket based communication
*/

/*
   The LD_LIBRARY_PATH needs to be set as follows:
      $JVMPATH (directory portion only)
      $JRE/lib/$ARCH
      $JRE/../lib/$ARCH
*/

#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>

#include <jni.h>

#define ARCH "i386"
#define JVM_DLL "libjvm.so"

typedef jint (JNICALL *CreateJavaVM_t)(JavaVM **pvm, void **env, void *args);
typedef jint (JNICALL *GetDefaultJavaVMInitArgs_t)(void *args);

typedef struct {
    CreateJavaVM_t CreateJavaVM;
    GetDefaultJavaVMInitArgs_t GetDefaultJavaVMInitArgs;
} InvocationFunctions;

const char * GetArch() {
    static char *arch = NULL;
    static char buf[12];
    
    if (arch) {
	return arch;
    }
    strcpy(buf, ARCH);
    arch = buf;
    return arch;
}

static jboolean GetJVMPath(const char *jrepath, const char *jvmtype,
                           char *jvmpath) {
    struct stat s;
    
    sprintf(jvmpath, "%s/lib/%s/%s/" JVM_DLL, jrepath, GetArch(), jvmtype);
    printf("%s\n", jvmpath);

    if (stat(jvmpath, &s) == 0) {
	return JNI_TRUE;
    } else {
	return JNI_FALSE;
    }
}

jboolean LoadJavaVM(const char *jvmpath, InvocationFunctions *ifn) {
    void *libjvm = NULL;

    libjvm = dlopen(jvmpath, RTLD_NOW + RTLD_GLOBAL);
    if (libjvm == NULL) {
        return JNI_FALSE;
    }

    ifn->CreateJavaVM = (CreateJavaVM_t) dlsym(libjvm, "JNI_CreateJavaVM");
    if (ifn->CreateJavaVM == NULL) 
        return JNI_FALSE;

    ifn->GetDefaultJavaVMInitArgs = (GetDefaultJavaVMInitArgs_t) dlsym(libjvm, "JNI_GetDefaultJavaVMInitArgs");
    if (ifn->GetDefaultJavaVMInitArgs == NULL) 
        return JNI_FALSE;
    return JNI_TRUE;
}

jboolean InitializeJVM(JavaVM **pvm, JNIEnv **penv, InvocationFunctions *ifn) {
    JavaVMInitArgs args;
    JavaVMOption options[1];
    jint numOptions = 1;
    jint r;
    char *def;
    char *s;
    char *check;

    s = getenv("CLASSPATH");
    if (s == NULL) {
       check = NULL;
    } else {
        check = strstr(s, "ffe.jar");
    }
    if (check == NULL){
        return JNI_FALSE;
    }
    
    def = (char *) malloc(strlen(s) + 40);
    sprintf(def, "-Djava.class.path=%s", s);
    options[0].optionString = def;

    memset(&args, 0, sizeof(args));
    args.version  = JNI_VERSION_1_2;
    args.nOptions = numOptions;
    args.options  = options;
    args.ignoreUnrecognized = JNI_FALSE;
    
    r = ifn->CreateJavaVM(pvm, (void **)penv, &args);
    return r == JNI_OK;
}

void createjvm_(int *locenv) {
    char jrepath[1024] = "/local/java/j2sdk1.4.2";
    char jvmpath[1024] = "libjvm.so";
    char *jvmtype = "client";
    InvocationFunctions ifn;
    char *classpath;
    char *ldpath;
    JavaVM *vm = 0;
    JNIEnv *env = 0;

    ifn.CreateJavaVM = 0;
    ifn.GetDefaultJavaVMInitArgs = 0;
    if (!LoadJavaVM(jvmpath, &ifn)) {
        *locenv = 0;
        return;
    }

    if (!InitializeJVM(&vm, &env, &ifn)) {
        *locenv = 0;
        return;
    }

    *locenv = (int) env;
    return;
}

void destroyjvm_(JNIEnv **loc){
    JNIEnv *env = *loc;
    JavaVM *jvm;
    (*env)->GetJavaVM(env, &jvm);
    (*jvm)->DestroyJavaVM(jvm);
}

void getmonitor_(JNIEnv **loc, jobject *obj){
    JNIEnv *env = *loc;
    (*env)->MonitorEnter(env, *obj);
}

void releasemonitor_(JNIEnv **loc, jobject *obj){
    JNIEnv *env = *loc;
    (*env)->MonitorExit(env, *obj);
}

void createserver_(JNIEnv **loc, jobject *system, jobject *server) {
    jclass class;
    jmethodID methodID;
    jobject temp;
    JNIEnv *env = *loc;

    class = (*env)->FindClass(env, "ffe/socket/FServer");
    if (class == 0) {
        printf(" Server not started: ffe.jar is not in the CLASSPATH\n\n");
        return;
    }
    methodID = (*env)->GetMethodID(env, class, "<init>",
               "(Lffe/socket/TinkerSystem;)V");
    temp = (*env)->NewObject(env, class, methodID, *system);
    methodID = (*env)->GetMethodID(env, class, "start", "()V");
    (*env)->CallVoidMethod(env, temp, methodID, "()V");
    *server =  (*env)->NewGlobalRef(env, temp);
    (*env)->DeleteLocalRef(env, temp);
    return;
}

void destroyserver_(JNIEnv **loc, jobject *server){
    jclass class;
    jmethodID methodID;
    JNIEnv *env = *loc;

    class = (*env)->GetObjectClass(env, *server);
    methodID = (*env)->GetMethodID(env, class, "stop", "()V");
    (*env)->CallVoidMethod(env, *server, methodID, "()V");

    methodID = (*env)->GetMethodID(env, class, "isAlive", "()Z");
    while ( (*env)->CallBooleanMethod(env, *server, methodID, "()Z")) { }
    (*env)->DeleteGlobalRef(env, *server);
    return;
}

void needupdate_(JNIEnv **loc, jobject *server, int *status){
    jclass class;
    jmethodID methodID;
    JNIEnv *env = *loc;
    jboolean ret;

    class = (*env)->GetObjectClass(env, *server);
    methodID = (*env)->GetMethodID(env, class, "needUpdate", "()Z");
    ret = (*env)->CallBooleanMethod(env, *server, methodID, "()Z");
    if (ret) *status = 1;
    else *status = 0;
    return;
}

void setupdated_(JNIEnv **loc, jobject *server){
    jclass class;
    jmethodID methodID;
    JNIEnv *env = *loc;

    class = (*env)->GetObjectClass(env, *server);
    methodID = (*env)->GetMethodID(env, class, "setUpdated", "()V");
    (*env)->CallVoidMethod(env, *server, methodID, "()V");
    return;
}

jstring char2jstring(JNIEnv *env, const char *str, int len) {
    jstring result;
    jbyteArray bytes = 0;
    jmethodID methodID;
    jclass class;

    if ((*env)->EnsureLocalCapacity(env, 2) < 0) {
        printf("Out of memory\n");
        exit(-1);
    }
    bytes = (*env)->NewByteArray(env, len);
    if (bytes != NULL) {
        (*env)->SetByteArrayRegion(env, bytes, 0, len,
                                   (jbyte *)str);
        class = (*env)->FindClass(env, "Ljava/lang/String;");
        methodID = (*env)->GetMethodID(env, class, "<init>", "([B)V");
        result = (*env)->NewObject(env, class, methodID, bytes);
        (*env)->DeleteLocalRef(env, bytes);
        if (result != NULL) return result;
    }
    printf("Problem creating java string for: %s\nLength: %i", str, len);
    exit(-1);
}

void createsystem_(JNIEnv **loc, jint *atoms, jint *keywords, jobject *system) {
    jclass class;
    jmethodID methodID;
    jobject temp;
    JNIEnv *env = *loc;

    class    = (*env)->FindClass(env, "ffe/socket/TinkerSystem");
    if (class == 0) {
        printf(" System not created: ffe.jar is not in the CLASSPATH\n\n");
        return;
    }
    methodID = (*env)->GetMethodID(env, class, "<init>", "(II)V");
    temp = (*env)->NewObject(env, class, methodID, *atoms, *keywords);
    *system  = (*env)->NewGlobalRef(env, temp);
    (*env)->DeleteLocalRef(env, temp);
    return;
}

void createupdate_(JNIEnv **loc, jint *n, jint *type, jint* amoeba,
                   jobject *update, jobject *server){
    jclass class;
    jmethodID methodID;
    jobject temp;
    jboolean bool;
    JNIEnv *env = *loc;

    if (amoeba == 0) bool = JNI_FALSE;
    else bool = JNI_TRUE;

    class    = (*env)->FindClass(env, "ffe/socket/TinkerUpdate");
    if (class == 0) {
        printf(" Update not created: ffe.jar is not in the CLASSPATH\n\n");
        return;
    }
    methodID = (*env)->GetMethodID(env, class, "<init>", "(IIZ)V");
    temp = (*env)->NewObject(env, class, methodID, *n, *type, bool);
    *update = (*env)->NewGlobalRef(env, temp);
    (*env)->DeleteLocalRef(env, temp);
    class = (*env)->GetObjectClass(env, *server);
    methodID = (*env)->GetMethodID(env,class,"loadUpdate",
                                   "(Lffe/socket/TinkerUpdate;)V");
    (*env)->CallObjectMethod(env, *server, methodID, *update);
    return;
}

void setcoordinates_(JNIEnv **loc, jobject *obj, jint *n,
                     jdouble *x, jdouble *y, jdouble *z) {
    jclass class;
    jfieldID fieldID;
    jobjectArray coords;
    jdoubleArray jx;
    JNIEnv *env = *loc;

    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "coordinates", "[[D");
    coords = (*env)->GetObjectField(env, *obj, fieldID);
    jx = (*env)->GetObjectArrayElement(env, coords, 0);
    (*env)->SetDoubleArrayRegion(env, jx, 0, *n, x);
    jx = (*env)->GetObjectArrayElement(env, coords, 1);
    (*env)->SetDoubleArrayRegion(env, jx, 0, *n, y);
    jx = (*env)->GetObjectArrayElement(env, coords, 2);
    (*env)->SetDoubleArrayRegion(env, jx, 0, *n, z);
}

void setconnectivity_(JNIEnv **loc, jobject *obj, jint *n,
                      jint *a, jint *b, jint *c, jint *d) {
    jclass class;
    jfieldID fieldID;
    jobjectArray connect;
    jintArray ci;
    JNIEnv *env = *loc;

    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "connectivity", "[[I");
    connect = (*env)->GetObjectField(env, *obj, fieldID);
    ci = (*env)->GetObjectArrayElement(env, connect, 0);
    (*env)->SetIntArrayRegion(env, ci, 0, *n, a);
    ci = (*env)->GetObjectArrayElement(env, connect, 1);
    (*env)->SetIntArrayRegion(env, ci, 0, *n, b);
    ci = (*env)->GetObjectArrayElement(env, connect, 2);
    (*env)->SetIntArrayRegion(env, ci, 0, *n, c);
    ci = (*env)->GetObjectArrayElement(env, connect, 3);
    (*env)->SetIntArrayRegion(env, ci, 0, *n, d);
}

/*
void setvdw_(JNIEnv **loc, jobject *obj, jint *n, jdouble *vdw) {
    jclass class;
    jfieldID fieldID;
    jdoubleArray jvdw;
    JNIEnv *env = *loc;

    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "vdw", "[D");
    jvdw = (*env)->GetObjectField(env, *obj, fieldID);
    (*env)->SetDoubleArrayRegion(env, jvdw, 0, *n, vdw);
}
*/

void setcharge_(JNIEnv **loc, jobject *obj, jint *n, jdouble *charge) {
    jclass class;
    jfieldID fieldID;
    jdoubleArray jcharge;
    JNIEnv *env = *loc;

    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "charge", "[D");
    jcharge = (*env)->GetObjectField(env, *obj, fieldID);
    (*env)->SetDoubleArrayRegion(env, jcharge, 0, *n, charge);
}

void setmass_(JNIEnv **loc, jobject *obj, jint *n, jdouble *mass) {
    jclass class;
    jfieldID fieldID;
    jdoubleArray jmass;
    JNIEnv *env = *loc;

    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "mass", "[D");
    jmass = (*env)->GetObjectField(env, *obj, fieldID);
    (*env)->SetDoubleArrayRegion(env, jmass, 0, *n, mass);
}

void setatomic_(JNIEnv **loc, jobject *obj, jint *n, jint *atomic) {
    jclass class;
    jfieldID fieldID;
    jintArray jatomic;
    JNIEnv *env = *loc;

    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "atomic", "[I");
    jatomic = (*env)->GetObjectField(env, *obj, fieldID);
    (*env)->SetIntArrayRegion(env, jatomic, 0, *n, atomic);
}

void setatomtypes_(JNIEnv **loc, jobject *obj, jint *n, jint *t) {
    jclass class;
    jfieldID fieldID;
    jintArray types;
    JNIEnv *env = *loc;

    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "types", "[I");
    types = (*env)->GetObjectField(env, *obj, fieldID);
    (*env)->SetIntArrayRegion(env, types, 0, *n, t);
}

void setname_(JNIEnv **loc, jobject *obj, jint *num, char *s, jint len) {
    jclass class;
    jfieldID fieldID;
    jstring string;
    jobjectArray name;
    JNIEnv *env = *loc;
    jint index;

    index = (*num) - 1;
    if (index < 0) {
        printf("Negative index into name array\n");
        exit(-1);
    }
    string = char2jstring(env, s, len);
    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "name",
                                 "[Ljava/lang/String;");
    name = (*env)->GetObjectField(env, *obj, fieldID);
    (*env)->SetObjectArrayElement(env, name, index, string);
}

void setstory_(JNIEnv **loc, jobject *obj, jint *num, char *s, jint len) {
    jclass class;
    jfieldID fieldID;
    jstring string;
    jobjectArray story;
    JNIEnv *env = *loc;
    jint index;

    index = (*num) - 1;
    if (index < 0) {
        printf("Negative index into story array\n");
        exit(-1);
    }
    string = char2jstring(env, s, len);
    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "story",
                                 "[Ljava/lang/String;");
    story = (*env)->GetObjectField(env, *obj, fieldID);
    (*env)->SetObjectArrayElement(env, story, index, string);
}

void setkeyword_(JNIEnv **loc, jobject *obj, jint *num,
                 char *keyword, jint len) {
    jclass class;
    jfieldID fieldID;
    jstring string;
    jobjectArray keywords;
    JNIEnv *env = *loc;
    jint index;

    index = (*num) - 1;
    if (index < 0) {
        printf("Negative index into keyword array\n");
        exit(0);
    }

    string = char2jstring(env, keyword, len);
    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "keywords",
                                 "[Ljava/lang/String;");
    keywords = (*env)->GetObjectField(env, *obj, fieldID);
    (*env)->SetObjectArrayElement(env, keywords, index, string);
}

void setforcefield_(JNIEnv **loc, jobject *obj, char *forcefield, jint len) {
    jclass class;
    jfieldID fieldID;
    jstring string;
    JNIEnv *env = *loc;

    string = char2jstring(env, forcefield, len);
    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "forcefield",
                                 "Ljava/lang/String;");
    (*env)->SetObjectField(env, *obj, fieldID, string);
}

void settime_(JNIEnv **loc, jobject *obj, jdouble *time) {
    jclass class;
    jfieldID fieldID;
    JNIEnv *env = *loc;

    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "time", "D");
    (*env)->SetDoubleField(env, *obj, fieldID, *time);
}

void setenergy_(JNIEnv **loc, jobject *obj, jdouble *e) {
    jclass class;
    jfieldID fieldID;
    JNIEnv *env = *loc;

    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "energy", "D");
    (*env)->SetDoubleField(env, *obj, fieldID, *e);
}

void setstep_(JNIEnv **loc, jobject *obj, jint *step) {
    jclass class;
    jfieldID fieldID;
    JNIEnv *env = *loc;

    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "step", "I");
    (*env)->SetIntField(env, *obj, fieldID, *step);
}

void settype_(JNIEnv **loc, jobject *obj, jint *type) {
    jclass class;
    jfieldID fieldID;
    JNIEnv *env = *loc;

    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "type", "I");
    (*env)->SetIntField(env, *obj, fieldID, *type);
}

void setfile_(JNIEnv **loc, jobject *obj, char *name, jint len) {
    jclass class;
    jfieldID fieldID;
    jstring string;
    JNIEnv *env = *loc;

    string = char2jstring(env, name, len);
    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "file",
                                 "Ljava/lang/String;");
    (*env)->SetObjectField(env, *obj, fieldID, string);
}

void setgradients_(JNIEnv **loc, jobject *obj, jint *n,
                   jdouble *x, jdouble *y, jdouble *z){
    jclass class;
    jfieldID fieldID;
    jobjectArray ind;
    jdoubleArray jx;
    JNIEnv *env = *loc;

    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "gradients", "[[D");
    ind = (*env)->GetObjectField(env, *obj, fieldID);
    jx = (*env)->GetObjectArrayElement(env, ind, 0);
    (*env)->SetDoubleArrayRegion(env, jx, 0, *n, x);
    jx = (*env)->GetObjectArrayElement(env, ind, 1);
    (*env)->SetDoubleArrayRegion(env, jx, 0, *n, y);
    jx = (*env)->GetObjectArrayElement(env, ind, 2);
    (*env)->SetDoubleArrayRegion(env, jx, 0, *n, z);
}

void setvelocity_(JNIEnv **loc, jobject *obj, jint *n,
                  jdouble *x, jdouble *y, jdouble *z){
    jclass class;
    jfieldID fieldID;
    jobjectArray ind;
    jdoubleArray jx;
    JNIEnv *env = *loc;

    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "velocity", "[[D");
    ind = (*env)->GetObjectField(env, *obj, fieldID);
    jx = (*env)->GetObjectArrayElement(env, ind, 0);
    (*env)->SetDoubleArrayRegion(env, jx, 0, *n, x);
    jx = (*env)->GetObjectArrayElement(env, ind, 1);
    (*env)->SetDoubleArrayRegion(env, jx, 0, *n, y);
    jx = (*env)->GetObjectArrayElement(env, ind, 2);
    (*env)->SetDoubleArrayRegion(env, jx, 0, *n, z);
}

void setacceleration_(JNIEnv **loc, jobject *obj, jint *n,
                      jdouble *x, jdouble *y, jdouble *z){
    jclass class;
    jfieldID fieldID;
    jobjectArray ind;
    jdoubleArray jx;
    JNIEnv *env = *loc;

    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "acceleration", "[[D");
    ind = (*env)->GetObjectField(env, *obj, fieldID);
    jx = (*env)->GetObjectArrayElement(env, ind, 0);
    (*env)->SetDoubleArrayRegion(env, jx, 0, *n, x);
    jx = (*env)->GetObjectArrayElement(env, ind, 1);
    (*env)->SetDoubleArrayRegion(env, jx, 0, *n, y);
    jx = (*env)->GetObjectArrayElement(env, ind, 2);
    (*env)->SetDoubleArrayRegion(env, jx, 0, *n, z);
}

void setinduced_(JNIEnv **loc, jobject *obj, jint *n,
                 jdouble *x, jdouble *y, jdouble *z){
    jclass class;
    jfieldID fieldID;
    jobjectArray ind;
    jdoubleArray jx;
    JNIEnv *env = *loc;

    class = (*env)->GetObjectClass(env, *obj);
    fieldID = (*env)->GetFieldID(env, class, "induced", "[[D");
    ind = (*env)->GetObjectField(env, *obj, fieldID);
    jx = (*env)->GetObjectArrayElement(env, ind, 0);
    (*env)->SetDoubleArrayRegion(env, jx, 0, *n, x);
    jx = (*env)->GetObjectArrayElement(env, ind, 1);
    (*env)->SetDoubleArrayRegion(env, jx, 0, *n, y);
    jx = (*env)->GetObjectArrayElement(env, ind, 2);
    (*env)->SetDoubleArrayRegion(env, jx, 0, *n, z);
}
