/* This file serves as an interface between TINKER FORTRAN code
   and FFE Java objects used for socket based communication */

#include <stdio.h>
#include <jni.h>
#include <stdlib.h>
#include <string.h>

#include <windows.h>
#include <sys/stat.h>

// Windows

typedef jint (JNICALL *CreateJavaVM_t)(JavaVM **pvm, void **env, void *args);
typedef jint (JNICALL *GetDefaultJavaVMInitArgs_t)(void *args);

#define DOTRELEASE  "1.4" /* Same for 1.4.1, 1.4.2 etc. */
#define JRE_KEY	    "Software\\JavaSoft\\Java Runtime Environment"
#define JVM_DLL "jvm.dll"

typedef struct {
    CreateJavaVM_t CreateJavaVM;
    GetDefaultJavaVMInitArgs_t GetDefaultJavaVMInitArgs;
} InvocationFunctions;

void CREATEJVM(int *locenv) {
        char jrepath[1024];
        char jvmpath[1024];
        char *jvmtype = "client";
        InvocationFunctions ifn;
        char *classpath;
        JavaVM *vm = 0;
        JNIEnv *env = 0;

        printf("Attempting to create JVM...\n");
        /* Find out where the JRE is that we will be using. */
        if (!GetPublicJREHome(jrepath, 1024)) {
                printf("Error: could not find Java 2 Runtime Environment.\n");
                  *locenv = 0;
                  return;
        }
        jvmpath[0] = '\0';
        if (!GetJVMPath(jrepath, jvmtype, jvmpath)) {
                printf("Error: no `%s' JVM at `%s'.", jvmtype,jvmpath);
                *locenv = 0;
                return;
        }
        ifn.CreateJavaVM = 0;
        ifn.GetDefaultJavaVMInitArgs = 0;
        if (!LoadJavaVM(jvmpath, &ifn)) {
                printf("Could not load JNI Invocation Functions.\n");
                *locenv = 0;
                return;
        }

        if (!InitializeJVM(&vm, &env, &ifn)) {
            printf("Could not create the Java virtual machine.\n");
                *locenv = 0;
                return;
        }

        *locenv = (int) env;
        return;
}

jboolean InitializeJVM(JavaVM **pvm, JNIEnv **penv, InvocationFunctions *ifn) {
    JavaVMInitArgs args;
    JavaVMOption options[1];
    jint numOptions = 1;
    jint r;
    char *def;
    char *s;
    char *check;

    // Set the classpath
    s = getenv("CLASSPATH");

    if (s == NULL) {
       check = NULL;
    } else {
       check = strstr(s, "ffe.jar");
    }
    if (check == NULL){
       return JNI_FALSE;
    }

    def = (char *) malloc(strlen(s) + 40);
    sprintf(def, "-Djava.class.path=%s", s);
    options[0].optionString = def;

    args.version  = JNI_VERSION_1_2;
    args.nOptions = numOptions;
    args.options  = options;
    args.ignoreUnrecognized = JNI_FALSE;
    r = ifn->CreateJavaVM(pvm, (void **)penv, &args);
    return r == JNI_OK;
}

jboolean GetJVMPath(const char *jrepath, const char *jvmtype,
       char *jvmpath) {
    struct stat s;
    sprintf(jvmpath, "%s\\bin\\%s\\" JVM_DLL, jrepath, jvmtype);
    if (stat(jvmpath, &s) == 0) {
        return JNI_TRUE;
    } else {
        return JNI_FALSE;
    }
}

jboolean LoadJavaVM(const char *jvmpath, InvocationFunctions *ifn) {
    HINSTANCE handle;

    /* Load the Java VM DLL */
    if ((handle = LoadLibrary(jvmpath)) == 0) {
    //    printf("Error loading: %s", (char *)jvmpath);
        return JNI_FALSE;
    }

    /* Now get the function addresses */
    ifn->CreateJavaVM = (void *)GetProcAddress(handle, "JNI_CreateJavaVM");
    ifn->GetDefaultJavaVMInitArgs =
                (void *)GetProcAddress(handle, "JNI_GetDefaultJavaVMInitArgs");

    if (ifn->CreateJavaVM == 0 || ifn->GetDefaultJavaVMInitArgs == 0) {
    //    printf("Error: can't find JNI interfaces in: %s", (char *)jvmpath);
        return JNI_FALSE;
    }
    return JNI_TRUE;
}

static jboolean GetStringFromRegistry(HKEY key, const char *name, char *buf, jint bufsize) {
    DWORD type, size;

    if (RegQueryValueEx(key, name, 0, &type, 0, &size) == 0 && type == REG_SZ
        && (size < (unsigned int)bufsize)) {
        if (RegQueryValueEx(key, name, 0, 0, buf, &size) == 0) {
            return JNI_TRUE;
        }
    }
    return JNI_FALSE;
}

jboolean GetPublicJREHome(char *buf, jint bufsize) {
    HKEY key, subkey;
    char version[1024];

    /* Find the current version of the JRE */
    if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, JRE_KEY, 0, KEY_READ, &key) != 0) {
    //    fprintf(stderr, "Error opening registry key '" JRE_KEY "'\n");
        return JNI_FALSE;
    }

    if (!GetStringFromRegistry(key, "CurrentVersion",
                               version, sizeof(version))) {
    //    fprintf(stderr, "Failed reading value of registry key:\n\t"
    //            JRE_KEY "\\CurrentVersion\n");
        RegCloseKey(key);
        return JNI_FALSE;
    }

    if (strcmp(version, DOTRELEASE) != 0) {
    //    fprintf(stderr, "Registry key '" JRE_KEY "\\CurrentVersion'\nhas "
    //            "value '%s', but '" DOTRELEASE "' is required.\n", version);
        RegCloseKey(key);
        return JNI_FALSE;
    }

    /* Find directory where the current version is installed. */
    if (RegOpenKeyEx(key, version, 0, KEY_READ, &subkey) != 0) {
    //    fprintf(stderr, "Error opening registry key '"
    //            JRE_KEY "\\%s'\n", version);
        RegCloseKey(key);
        return JNI_FALSE;
    }

    if (!GetStringFromRegistry(subkey, "JavaHome", buf, bufsize)) {
    //    fprintf(stderr, "Failed reading value of registry key:\n\t"
    //            JRE_KEY "\\%s\\JavaHome\n", version);
        RegCloseKey(key);
        RegCloseKey(subkey);
        return JNI_FALSE;
    }

    RegCloseKey(key);
    RegCloseKey(subkey);
    return JNI_TRUE;
}

void DESTROYJVM(JNIEnv **loc){
        JNIEnv *env = *loc;
	JavaVM *jvm;
	(*env)->GetJavaVM(env, &jvm);	
	(*jvm)->DestroyJavaVM(jvm);
}

void GETMONITOR(JNIEnv **loc, jobject *obj){ 
        JNIEnv *env = *loc;
	(*env)->MonitorEnter(env, *obj);
}

void RELEASEMONITOR(JNIEnv **loc, jobject *obj){
        JNIEnv *env = *loc;
  	(*env)->MonitorExit(env, *obj);
}

void CREATESERVER(JNIEnv **loc, jobject *system, jobject *server) {
	jclass class;
	jmethodID methodID;
        jobject temp;
        JNIEnv *env = *loc;

	// Find the Java ffe.socket.FServer class
	class = (*env)->FindClass(env, "ffe/socket/FServer");
	if (class == 0) {
	   printf(" Server not started: ffe.jar is not in the CLASSPATH\n\n");
           return;
	}
	methodID = (*env)->GetMethodID(env, class, "<init>", 
		"(Lffe/socket/TinkerSystem;)V");
	temp =  (*env)->NewObject(env, class, methodID, *system);
	methodID = (*env)->GetMethodID(env, class, "start", "()V");
	(*env)->CallVoidMethod(env, temp, methodID, "()V");
	*server =  (*env)->NewGlobalRef(env, temp); 
	(*env)->DeleteLocalRef(env, temp);
	return;
}

//void DESTROYSERVER(JNIEnv **loc, jobject *server, int *ffe){
void DESTROYSERVER(JNIEnv **loc, jobject *server){
	jclass class;
	jmethodID methodID;
        JNIEnv *env = *loc;
        jboolean jffe = JNI_FALSE;
       
	// If FFE launched the TINKER EXE, ffe will be one (true)
//	if (ffe) {
//	jffe = JNI_TRUE;
//	} 

	// Tell the server to close down
	class = (*env)->GetObjectClass(env, *server);
	methodID = (*env)->GetMethodID(env, class, "stop", "()V");
	(*env)->CallVoidMethod(env, *server, methodID, "()V");
//	methodID = (*env)->GetMethodID(env, class, "stop", "(Z)V");
//	(*env)->CallVoidMethod(env, *server, methodID, "(Z)V", jffe);
	
	// Wait while it closes any clients
	methodID = (*env)->GetMethodID(env, class, "isAlive", "()Z");
	while ( (*env)->CallBooleanMethod(env, *server, methodID, "()Z")) { 

        }
        (*env)->DeleteGlobalRef(env, *server);
        return;
}

void NEEDUPDATE(JNIEnv **loc, jobject *server, int *status){
	jclass class;
	jmethodID methodID;
        JNIEnv *env = *loc;
	jboolean ret;

	class = (*env)->GetObjectClass(env, *server);
	methodID = (*env)->GetMethodID(env, class, "needUpdate", "()Z");
	ret = (*env)->CallBooleanMethod(env, *server, methodID, "()Z");
        if (ret) *status = 1;
	else *status = 0;
        return;
}

void SETUPDATED(JNIEnv **loc, jobject *server){
	jclass class;
	jmethodID methodID;
        JNIEnv *env = *loc;

	class = (*env)->GetObjectClass(env, *server);
	methodID = (*env)->GetMethodID(env, class, "setUpdated", "()V");
	(*env)->CallVoidMethod(env, *server, methodID, "()V");
        return;
}

jstring char2jstring(JNIEnv *env, const char *str, int len) {
     jstring result;
     jbyteArray bytes = 0;
     jmethodID methodID;
     jclass class;

     if ((*env)->EnsureLocalCapacity(env, 2) < 0) {
         printf("Out of memory\n");
         exit(-1);
     }
     bytes = (*env)->NewByteArray(env, len);
     if (bytes != NULL) {
         (*env)->SetByteArrayRegion(env, bytes, 0, len,
                                    (jbyte *)str);
	 class = (*env)->FindClass(env, "Ljava/lang/String;");
         methodID = (*env)->GetMethodID(env, class, "<init>", "([B)V");
         result = (*env)->NewObject(env, class, methodID, bytes);
         (*env)->DeleteLocalRef(env, bytes);
         if (result != NULL) return result;
     } /* else fall through */
     printf("Problem creating java string for: %s\nLength: %i", str, len);
     exit(-1); 
}

void CREATESYSTEM(JNIEnv **loc, jint *atoms, jint *keywords, jobject *system) {
	jclass class;
	jmethodID methodID;
        jobject temp;
        JNIEnv *env = *loc;

	class    = (*env)->FindClass(env, "ffe/socket/TinkerSystem");
        if (class == 0) {
	   printf(" System not created: ffe.jar is not in the CLASSPATH\n\n");
           return;
	}
	methodID = (*env)->GetMethodID(env, class, "<init>", "(II)V");
	temp     = (*env)->NewObject(env, class, methodID, *atoms, *keywords);
        *system  = (*env)->NewGlobalRef(env, temp); 
        (*env)->DeleteLocalRef(env, temp);
	return;
}

void CREATEUPDATE(JNIEnv **loc, jint *n, jint *type, jint* amoeba, 
		  jobject *update, jobject *server){ 
	jclass class;
	jmethodID methodID;
        jobject temp;
        jboolean bool;
        JNIEnv *env = *loc;

        if (amoeba == 0) bool = JNI_FALSE;
        else bool = JNI_TRUE;

	class    = (*env)->FindClass(env, "ffe/socket/TinkerUpdate");
        if (class == 0) {
	   printf(" Update not created: ffe.jar is not in the CLASSPATH\n\n");
           return;
	}
	methodID = (*env)->GetMethodID(env, class, "<init>", "(IIZ)V");
	temp     = (*env)->NewObject(env, class, methodID, *n, *type, bool);
        *update  = (*env)->NewGlobalRef(env, temp); 
        (*env)->DeleteLocalRef(env, temp);
	class    = (*env)->GetObjectClass(env, *server);
	methodID = (*env)->GetMethodID(env,class,"loadUpdate",
			  "(Lffe/socket/TinkerUpdate;)V");
        (*env)->CallObjectMethod(env, *server, methodID, *update); 
	return;
}

void SETCOORDINATES(JNIEnv **loc, jobject *obj, jint *n, 
		   jdouble *x, jdouble *y, jdouble *z) {
        jclass class;
	jfieldID fieldID;
        jobjectArray coords;
        jdoubleArray jx; 
        JNIEnv *env = *loc;

	class = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "coordinates", "[[D");
	coords = (*env)->GetObjectField(env, *obj, fieldID);
        jx = (*env)->GetObjectArrayElement(env, coords, 0);
        (*env)->SetDoubleArrayRegion(env, jx, 0, *n, x);
        jx = (*env)->GetObjectArrayElement(env, coords, 1);
        (*env)->SetDoubleArrayRegion(env, jx, 0, *n, y);
        jx = (*env)->GetObjectArrayElement(env, coords, 2);
        (*env)->SetDoubleArrayRegion(env, jx, 0, *n, z);
}

void SETCONNECTIVITY(JNIEnv **loc, jobject *obj, jint *n, 
		   jint *a, jint *b, jint *c, jint *d) {
        jclass class;
	jfieldID fieldID;
        jobjectArray connect;
        jintArray ci; 
        JNIEnv *env = *loc;

	class = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "connectivity", "[[I");
	connect = (*env)->GetObjectField(env, *obj, fieldID);
        ci = (*env)->GetObjectArrayElement(env, connect, 0);
        (*env)->SetIntArrayRegion(env, ci, 0, *n, a);
        ci = (*env)->GetObjectArrayElement(env, connect, 1);
        (*env)->SetIntArrayRegion(env, ci, 0, *n, b);
        ci = (*env)->GetObjectArrayElement(env, connect, 2);
        (*env)->SetIntArrayRegion(env, ci, 0, *n, c);
        ci = (*env)->GetObjectArrayElement(env, connect, 3);
        (*env)->SetIntArrayRegion(env, ci, 0, *n, d);
}

/*
void SETVDW(JNIEnv **loc, jobject *obj, jint *n, jdouble *vdw) {
        jclass class;
	jfieldID fieldID;
        jdoubleArray jvdw;
        JNIEnv *env = *loc;

	class = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "vdw", "[D");
	jvdw = (*env)->GetObjectField(env, *obj, fieldID);
        (*env)->SetDoubleArrayRegion(env, jvdw, 0, *n, vdw);
}*/

void SETCHARGE(JNIEnv **loc, jobject *obj, jint *n, jdouble *charge) {
        jclass class;
	jfieldID fieldID;
        jdoubleArray jcharge;
        JNIEnv *env = *loc;

	class = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "charge", "[D");
	jcharge = (*env)->GetObjectField(env, *obj, fieldID);
        (*env)->SetDoubleArrayRegion(env, jcharge, 0, *n, charge);
}

void SETMASS(JNIEnv **loc, jobject *obj, jint *n, jdouble *mass) {
        jclass class;
	jfieldID fieldID;
        jdoubleArray jmass;
        JNIEnv *env = *loc;

	class = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "mass", "[D");
	jmass = (*env)->GetObjectField(env, *obj, fieldID);
        (*env)->SetDoubleArrayRegion(env, jmass, 0, *n, mass);
}

void SETATOMIC(JNIEnv **loc, jobject *obj, jint *n, jint *atomic) {
        jclass class;
	jfieldID fieldID;
        jintArray jatomic;
        JNIEnv *env = *loc;

	class = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "atomic", "[I");
	jatomic = (*env)->GetObjectField(env, *obj, fieldID);
        (*env)->SetIntArrayRegion(env, jatomic, 0, *n, atomic);
}

void SETATOMTYPES(JNIEnv **loc, jobject *obj, jint *n, jint *t) {
        jclass class;
	jfieldID fieldID;
        jintArray types;
        JNIEnv *env = *loc;

	class = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "types", "[I");
	types = (*env)->GetObjectField(env, *obj, fieldID);
        (*env)->SetIntArrayRegion(env, types, 0, *n, t);
}

void SETNAME(JNIEnv **loc, jobject *obj, jint *num, 
		char *s, jint len) {
        jclass class;
	jfieldID fieldID;
        jstring string;
        jobjectArray name;
        JNIEnv *env = *loc;
        jint index;

	index = (*num) - 1;
        if (index < 0) {
		printf("Negative index into name array\n");
                exit(-1);
	}
 	string = char2jstring(env, s, len);	
	class = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "name", 
 				"[Ljava/lang/String;");
	name = (*env)->GetObjectField(env, *obj, fieldID);
        (*env)->SetObjectArrayElement(env, name, index, string);
}

void SETSTORY(JNIEnv **loc, jobject *obj, jint *num, char *s, jint len) {
        jclass class;
	jfieldID fieldID;
        jstring string;
        jobjectArray story;
        JNIEnv *env = *loc;
        jint index;

	index = (*num) - 1;
        if (index < 0) {
		printf("Negative index into story array\n");
	 	exit(-1);	
	}
 	string = char2jstring(env, s, len);	
	class = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "story", 
 				"[Ljava/lang/String;");
	story = (*env)->GetObjectField(env, *obj, fieldID);
        (*env)->SetObjectArrayElement(env, story, index, string);
}

void SETKEYWORD(JNIEnv **loc, jobject *obj, jint *num, 
		char *keyword, jint len) {
        jclass class;
	jfieldID fieldID;
        jstring string;
        jobjectArray keywords;
        JNIEnv *env = *loc;
        jint index;

	index = (*num) - 1;
        if (index < 0) {
		printf("Negative index into keyword array\n");
		exit(0);
	}

 	string = char2jstring(env, keyword, len);	
	class = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "keywords", 
 				"[Ljava/lang/String;");
	keywords = (*env)->GetObjectField(env, *obj, fieldID);
        (*env)->SetObjectArrayElement(env, keywords, index, string);
}

void SETFORCEFIELD(JNIEnv **loc, jobject *obj, char *forcefield, jint len) { 
        jclass class;
	jfieldID fieldID;
        jstring string;
        JNIEnv *env = *loc;

 	string = char2jstring(env, forcefield, len);	
	class   = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "forcefield", 
				    "Ljava/lang/String;");
	(*env)->SetObjectField(env, *obj, fieldID, string);
}

void SETTIME(JNIEnv **loc, jobject *obj, jdouble *time) { 
        jclass class;
	jfieldID fieldID;
        JNIEnv *env = *loc;

	class   = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "time", "D");
	(*env)->SetDoubleField(env, *obj, fieldID, *time);
}

void SETENERGY(JNIEnv **loc, jobject *obj, jdouble *e) { 
        jclass class;
	jfieldID fieldID;
        JNIEnv *env = *loc;

	class   = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "energy", "D");
	(*env)->SetDoubleField(env, *obj, fieldID, *e);
}

void SETSTEP(JNIEnv **loc, jobject *obj, jint *step) { 
        jclass class;
	jfieldID fieldID;
        JNIEnv *env = *loc;

	class   = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "step", "I");
	(*env)->SetIntField(env, *obj, fieldID, *step);
}

void SETTYPE(JNIEnv **loc, jobject *obj, jint *type) { 
        jclass class;
	jfieldID fieldID;
        JNIEnv *env = *loc;

	class   = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "type", "I");
	(*env)->SetIntField(env, *obj, fieldID, *type);
}

void SETFILE(JNIEnv **loc, jobject *obj, char *name, jint len) { 
        jclass class;
	jfieldID fieldID;
        jstring string;
        JNIEnv *env = *loc;

 	string = char2jstring(env, name, len);	
	class   = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "file", 
				    "Ljava/lang/String;");
	(*env)->SetObjectField(env, *obj, fieldID, string);
}

void SETGRADIENTS(JNIEnv **loc, jobject *obj, jint *n, 
		  jdouble *x, jdouble *y, jdouble *z){
        jclass class;
	jfieldID fieldID;
        jobjectArray ind;
        jdoubleArray jx; 
        JNIEnv *env = *loc;

	class = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "gradients", "[[D");
	ind = (*env)->GetObjectField(env, *obj, fieldID);
        jx = (*env)->GetObjectArrayElement(env, ind, 0);
        (*env)->SetDoubleArrayRegion(env, jx, 0, *n, x);
        jx = (*env)->GetObjectArrayElement(env, ind, 1);
        (*env)->SetDoubleArrayRegion(env, jx, 0, *n, y);
        jx = (*env)->GetObjectArrayElement(env, ind, 2);
        (*env)->SetDoubleArrayRegion(env, jx, 0, *n, z);
}

void SETVELOCITY(JNIEnv **loc, jobject *obj, jint *n, 
		 jdouble *x, jdouble *y, jdouble *z){
        jclass class;
	jfieldID fieldID;
        jobjectArray ind;
        jdoubleArray jx; 
        JNIEnv *env = *loc;

	class = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "velocity", "[[D");
	ind = (*env)->GetObjectField(env, *obj, fieldID);
        jx = (*env)->GetObjectArrayElement(env, ind, 0);
        (*env)->SetDoubleArrayRegion(env, jx, 0, *n, x);
        jx = (*env)->GetObjectArrayElement(env, ind, 1);
        (*env)->SetDoubleArrayRegion(env, jx, 0, *n, y);
        jx = (*env)->GetObjectArrayElement(env, ind, 2);
        (*env)->SetDoubleArrayRegion(env, jx, 0, *n, z);
}

void SETACCELERATION(JNIEnv **loc, jobject *obj, jint *n, 
		 jdouble *x, jdouble *y, jdouble *z){
        jclass class;
	jfieldID fieldID;
        jobjectArray ind;
        jdoubleArray jx; 
        JNIEnv *env = *loc;

	class = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "acceleration", "[[D");
	ind = (*env)->GetObjectField(env, *obj, fieldID);
        jx = (*env)->GetObjectArrayElement(env, ind, 0);
        (*env)->SetDoubleArrayRegion(env, jx, 0, *n, x);
        jx = (*env)->GetObjectArrayElement(env, ind, 1);
        (*env)->SetDoubleArrayRegion(env, jx, 0, *n, y);
        jx = (*env)->GetObjectArrayElement(env, ind, 2);
        (*env)->SetDoubleArrayRegion(env, jx, 0, *n, z);
}

void SETINDUCED(JNIEnv **loc, jobject *obj, jint *n, 
		jdouble *x, jdouble *y, jdouble *z){
        jclass class;
	jfieldID fieldID;
        jobjectArray ind;
        jdoubleArray jx; 
        JNIEnv *env = *loc;

	class = (*env)->GetObjectClass(env, *obj);
	fieldID = (*env)->GetFieldID(env, class, "induced", "[[D");
	ind = (*env)->GetObjectField(env, *obj, fieldID);
        jx = (*env)->GetObjectArrayElement(env, ind, 0);
        (*env)->SetDoubleArrayRegion(env, jx, 0, *n, x);
        jx = (*env)->GetObjectArrayElement(env, ind, 1);
        (*env)->SetDoubleArrayRegion(env, jx, 0, *n, y);
        jx = (*env)->GetObjectArrayElement(env, ind, 2);
        (*env)->SetDoubleArrayRegion(env, jx, 0, *n, z);
}
